from django.db import models
from django.utils import timezone
from django.db.models import Q

class CostCenterQuerySet(models.QuerySet):

    def active(self):
        now = timezone.now()
        return self.filter(
            (Q(start_date=None) & Q(end_date=None)) |
            (Q(start_date__lte=now) and Q(end_date=None)) |
            (Q(start_date=None) and Q(end_date__gte=now)) |
            (Q(start_date__lte=now) and Q(end_date__gte=now))
        )


from django.db import models
from .project_queryset import ProjectQuerySet

class Project(models.Model):
    """
    Represents Project in the system
    """
    name = models.CharField('Name', max_length=200) #: Name
    code = models.CharField('Code', max_length=200, unique=True) #: Number

    # Grant section
    start_date   = models.DateField('Starting date', blank=True, null=True,
                                    help_text='Starting date of the grant')
    end_date     = models.DateField('Ending date', blank=True, null=True,
                                    help_text='Ending date of the grant')
    total_amount = models.DecimalField('Total amount', blank=True, null=True, max_digits=11, decimal_places=2,
                                       help_text='Total amount of the grant')
    overheads    = models.DecimalField('Overhead', blank=True, null=True, max_digits=11, decimal_places=2,
                                      help_text='Overhead amount of the grant')
    funding      = models.DecimalField('Funding', blank=True, null=True, max_digits=11, decimal_places=2,
                                      help_text='Funding amount of the grant')

    responsible = models.ForeignKey('people.Person', on_delete=models.CASCADE)
    costcenter  = models.ForeignKey('CostCenter', verbose_name='Cost Center', on_delete=models.CASCADE)
    currency    = models.ForeignKey('common.Currency', verbose_name='Currency', blank=True, null=True, on_delete=models.CASCADE)
    grant       = models.ForeignKey('Grant', verbose_name='Grant', blank=True, null=True, on_delete=models.CASCADE)

    objects = ProjectQuerySet.as_manager()

    class Meta:
        ordering = ['code', 'costcenter']
        verbose_name = "Finance Project"
        verbose_name_plural = "Finance Projects"

    def __str__(self):
        costcenter = self.costcenter
        if costcenter!=None:
            return costcenter.code + "-" + self.code + ": " + costcenter.name + "-" + self.name
        else:
            return "No FinanceCostCenter associated - " + self.name + " - " + self.code

    # Custom methods

    def nationality(self):
        if self.grant is None:
            return ""
        return str(self.grant.nationality)

    def domain(self):
        if self.grant is None:
            return ""
        return str(self.grant.domain)

    def grantor(self):
        if self.grant is None:
            return ""
        return str(self.grant.grantor)



    @staticmethod
    def autocomplete_search_fields():
        return (
            'name__icontains',
            'code__icontains',
            'costcenter__name__icontains',
            'costcenter__code__icontains'
        )



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from django.db.models import Q
from django.template.defaultfilters import pluralize

from humanresources.utils import send_mail


class Command(BaseCommand):
    help = "Inspect all Orders in the DB for inconsistencies."

    def add_arguments(self, parser):
        parser.add_argument("--responsible", type=str, nargs="+")
        parser.add_argument("--notify", action="store_true")

    def handle(self, *args, **options):
        User = apps.get_model("auth", "User")
        Order = apps.get_model("finance", "Order")

        users_to_notify = []
        optional_filters = Q()

        if options["responsible"]:
            for username in options["responsible"]:
                try:
                    user = User.objects.get(username=username)
                except User.DoesNotExist:
                    raise CommandError("User '%s' not found" % username)
                users_to_notify.append(user)
                optional_filters.add(Q(responsible=user), Q.OR)

        # apply initial filters
        orders = Order.objects.filter(optional_filters)

        self.stdout.write(f"Found {orders.count()} Orders")

        if options["responsible"]:
            self.stdout.write(
                f"Found {orders.count()} Orders belonging to user{pluralize(options['responsible'])} {', '.join([repr(u) for u in options['responsible']])}"
            )

        orders_missing_supplier = orders.filter(supplier__isnull=True)
        if orders_missing_supplier:
            self.stdout.write(
                self.style.WARNING(
                    f"{orders_missing_supplier.count()} Orders are missing a Supplier"
                )
            )

        all_warnings = orders_missing_supplier

        if all_warnings and options["responsible"] and options["notify"]:
            # Send a notification email to each user with inconsistent data records

            subject = "Data quality report"

            for user in users_to_notify:
                if not all_warnings.filter(responsible=user):
                    continue

                orders_missing_supplier = all_warnings.filter(
                    responsible=user
                ).order_by('order_reqdate')

                recipient_list = [user.email]

                message_context = {
                    "base_url": settings.BASE_URL,
                    "orders_missing_supplier": orders_missing_supplier,
                }

                send_mail(subject, recipient_list, message_context)

                self.stdout.write(self.style.SUCCESS(f"Report sent to {user.email}"))
        elif all_warnings and options["responsible"] and not options["notify"]:
            self.stdout.write(
                "Pass the optional argument '--notify' to send an email to the users above"
            )
        elif all_warnings and not options["responsible"] and options["notify"]:
            users_to_warn = [
                User.objects.get(id=user_id)
                for user_id in all_warnings.values_list("responsible", flat=True)
                .distinct()
                .order_by("responsible__username")
            ]
            for user in users_to_warn:
                user_orders = all_warnings.filter(responsible=user)
                print(f"  {user_orders.count():3d} belonging to user '{user}'")
            self.stdout.write(
                self.style.ERROR(
                    "Mass notification is disabled, please select the users you wish to notify"
                )
            )
            self.stdout.write(
                "Pass the argument '--reponsible <user1> <user2>' to send an email to the selected users"
            )
        else:
            self.stdout.write("Nothing to report.")




from django.conf import settings
from django.core.exceptions import ValidationError

from confapp import conf

from pyforms.controls import ControlButton
from pyforms_web.organizers import no_columns
from pyforms_web.widgets.django import ModelAdminWidget
from pyforms_web.widgets.django import ModelFormWidget

from finance.models import Budget
from finance.models import BudgetDoc


class ImportBudgetFormWidget(ModelFormWidget):
    TITLE = 'Import Budget'

    MODEL = BudgetDoc

    CREATE_BTN_LABEL = '<i class="upload icon"></i>Import'
    HAS_CANCEL_BTN_ON_ADD = False

    FIELDSETS = ['document', 'year']

    LAYOUT_POSITION = conf.ORQUESTRA_NEW_WINDOW

    AUTHORIZED_GROUPS = ['superuser', settings.PROFILE_EDIT_BUDGET]

    def save_form_event(self, obj, **kwargs):
        obj = self.update_object_fields(obj)

        # Custom validation
        #   need to run the validators individually or else Django
        #   concatenates all erros together
        try:
            obj.clean_fields()
            obj.clean()
            obj.validate_unique()

        except ValidationError as e:
            errors = []  # a list of messages to display using Pyforms alert()

            for error_message in e:
                if isinstance(error_message, tuple):
                    field_name, messages = error_message

                    pyforms_field = getattr(self, field_name)
                    pyforms_field.error = True

                    if len(messages) == 1:
                        field_messages = f'{messages[0]}'
                    else:
                        field_messages = ''
                        for message in messages:
                            field_messages += f'<li>{message}</li>'
                        field_messages = f'<ul>{field_messages}</ul>'

                    html = f'<b>{pyforms_field.label}:</b> {field_messages}'

                else:
                    html = f'{error_message}'

                errors.append(html)

            self.alert(errors)

        else:
            obj.save(**kwargs)
            self.parent.success("Budgets updated successfully!")
            self.close()


class BudgetFormAdmin(ModelFormWidget):

    MODEL = Budget
    TITLE = 'Suppliers'

    FIELDSETS = [
        'expensecode',
        no_columns('year', 'amount'),
    ]

    AUTHORIZED_GROUPS = ['superuser']


class BudgetAdminWidget(ModelAdminWidget):

    UID = 'budgets'
    MODEL = Budget

    TITLE = 'Budgets'

    LIST_DISPLAY = [
        'year',
        'amount',
        'expensecode',
    ]

    LIST_FILTER = [
        'year',
        'expensecode__project',
        'expensecode__number',
        'expensecode__project__costcenter',
    ]

    EDITFORM_CLASS = BudgetFormAdmin

    ###########################################################################
    # ORQUESTRA CONFIGURATION

    LAYOUT_POSITION = conf.ORQUESTRA_HOME
    ORQUESTRA_MENU = 'left>FinancesDashboardWidget'
    ORQUESTRA_MENU_ORDER = 0
    # ORQUESTRA_MENU_ICON = 'dollar'
    AUTHORIZED_GROUPS = ['superuser', settings.PROFILE_EDIT_BUDGET]
    ###########################################################################

    def __init__(self, *args, **kwargs):

        self._import_btn = ControlButton(
            '<i class="upload icon"></i>Import',
            default=self.__import_evt,
            label_visible=False,
            css='basic blue',
            helptext='Import budget list from XLS file',
        )

        super().__init__(*args, **kwargs)

    def get_toolbar_buttons(self, *args, **kwargs):
        add_btn = super().get_toolbar_buttons(*args, **kwargs)
        return no_columns(add_btn, '_import_btn')

    def __import_evt(self):
        ImportBudgetFormWidget(parent_win=self)



from django.db import models

class Grantdomain(models.Model):
    """
    Represents a Grant domain in the system
    Example: public, private
    """
    domain = models.CharField('Grant domain', max_length=200)

    class Meta:
        verbose_name = "Grant domain"
        verbose_name_plural = "Grant domains"

    def __str__(self):
        return self.domain





from confapp import conf
from pyforms_web.widgets.django import ModelAdminWidget

from finance.models import FinancingAgency

class FinancingAgencyAdminApp(ModelAdminWidget):

    UID   = 'financing-agency'
    TITLE = 'Financing agency'
    MODEL = FinancingAgency

    LIST_DISPLAY = ['name']
    FIELDSETS = ['name']
    SEARCH_FIELDS = ['name__icontains']

    ########################################################
    #### ORQUESTRA CONFIGURATION ###########################
    ########################################################
    LAYOUT_POSITION      = conf.ORQUESTRA_HOME
    ORQUESTRA_MENU       = 'left>FinancesDashboardWidget'
    ORQUESTRA_MENU_ORDER = 0
    ORQUESTRA_MENU_ICON  = 'flag'
    ########################################################

    AUTHORIZED_GROUPS = ['superuser']
    


"""
from finance import views
try:
    # django 2.0
    from django.urls import path
except:
    # django 1.6
    from django.conf.urls import url as path


urlpatterns = [
    path('finance/finance/budgetslist/', views.budgetslist, name='budgetslist'),
    path('finance/budgetslist/finance/importbudgets/<str:docname>/', views.budgetslist, name='budgetslist'),
    path('finance/importbudgets/<int:docname>/', views.importbudgets,                    name='importbudgets'),
    path('finance/importbudgets/<str:docname>/', views.importbudgets,                    name='importbudgets'),
    path('export/orders_from_project/<int:project_id>/', views.orders_from_project),
]
"""


from django.db import models

class FinancingAgency(models.Model):
    """
    Represents a Grant grantor in the system
    Example: Bial, FCT
    """

    name = models.CharField('Name', max_length=200)  #: Name

    class Meta:
        verbose_name = "Financing agency"
        verbose_name_plural = "Financing agencies"
        ordering = ['name']

    def __str__(self):
        return self.name


import datetime
from decimal import Decimal
import xlrd

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):

    help = """
    Import a XLS file containing a list of Orders.

    Supports the format used by the Teaching Lab as of December 2018,
    with the following columns in the given order:

        - Description
        - Requester Name
        - Supplier
        - Amount (net)
        - Requisition Number
        - Req. date
        - Purchase Orders
        - Expense_Codes
        - Centro Responsabilidade
        - Project
        - Group
        - Payment Method
        - Notes
    """

    def add_arguments(self, parser):
        parser.add_argument("filename", type=str)
        parser.add_argument(
            "--no-header", action="store_false", help="disable header row"
        )

    def handle(self, *args, **options):

        errors = []

        workbook = xlrd.open_workbook(
            filename=options["filename"], encoding_override="utf-8", on_demand=True
        )
        worksheet = workbook.sheet_by_index(0)

        # start by validating the file structure
        #   - number, name and order of columns
        #   - empty cells

        expected_column_labels = [
            "Description",
            "Requester Name",
            "Supplier",
            "Amount (net)",
            "Requisition Number",
            "Req. date",
            "Purchase Orders",
            "Expense_Codes",
            "Centro Responsabilidade",
            "Project",
            "Group",
            "Payment Method",
            "Notes",
        ]
        expected_column_number = len(expected_column_labels)

        if worksheet.ncols != expected_column_number:
            raise CommandError(
                "Found %(found)s columns, expected exactly %(expected)s"
                % {"found": worksheet.ncols, "expected": expected_column_number}
            )

        for col, expected in zip(range(worksheet.ncols), expected_column_labels):
            if worksheet.cell_value(0, col) != expected:
                errors.append('Column "%s" not found' % expected)

        if errors:
            raise CommandError(errors)

        for col in range(worksheet.ncols):
            if col in [0, 2, 4, 5, 6, 7, 8, 9, 11, 12]:
                # ignore empty cells in columns for not mandatory fields:
                continue

            col_values = worksheet.col_values(col, start_rowx=2)
            empty_cells_row_indices = [
                i for i, value in enumerate(col_values) if value == ""
            ]
            for row in empty_cells_row_indices:
                cell = xlrd.formula.cellname(row, col)
                errors.append("Missing value in cell %s" % cell)

        if errors:
            raise CommandError(errors)

        # validate columns values
        # new entries must not have duplicated requisition numbers nor
        # purchase order number

        User = apps.get_model("auth", "User")
        AuthGroup = apps.get_model("auth", "Group")
        Currency = apps.get_model("finance", "Currency")
        OrderExpenseCode = apps.get_model("finance", "OrderExpenseCode")
        ExpenseCode = apps.get_model("finance", "ExpenseCode")
        FinanceProject = apps.get_model("finance", "FinanceProject")
        FinanceCostCenter = apps.get_model("finance", "FinanceCostCenter")
        Order = apps.get_model("finance", "Order")
        Supplier = apps.get_model("finance", "Supplier")

        self.stdout.write("Validating document...")

        imported_orders = []

        for row in range(1, worksheet.nrows):
            values = worksheet.row_values(row)

            # replace en-dash for single dash across all str values extracted
            values = [
                value.replace("\u2013", "-") if isinstance(value, str) else value
                for value in values
            ]

            values_raw = {
                col_name: value
                for col_name, value in zip(expected_column_labels, values)
            }

            # description --> TextField
            column_id = 0
            description = values_raw[expected_column_labels[column_id]]
            assert isinstance(description, str)

            # requester name --> CharField
            column_id = 1
            requester_name = values_raw[expected_column_labels[column_id]]
            assert "\n" not in requester_name, "Invalid requester name"
            assert isinstance(requester_name, str)

            # finance --> ForeignKey to Supplier
            column_id = 2
            supplier_name = values_raw[expected_column_labels[column_id]]
            assert "\n" not in supplier_name, "Invalid finance name"
            assert isinstance(supplier_name, str)
            if "_reimbursement" in supplier_name.lower():
                supplier_name = "Geral Reembolsos"
            if supplier_name.lower() == "":
                supplier = None
            else:
                try:
                    supplier = Supplier.objects.get(supplier_name=supplier_name.strip())
                except Supplier.DoesNotExist:
                    # Ignore missing suppliers in the DB
                    # Generate the report and warn the users to fill in
                    # the missing entries
                    supplier = None

                    msg = " ".join(
                        [
                            self.style.WARNING(f"Row {row+1:3d}:"),
                            f"Verify finance name '{supplier_name}'",
                        ]
                    )
                    self.stdout.write(msg)
                else:
                    assert isinstance(supplier, Supplier)

            # ammount --> DecimalField
            column_id = 3
            ammount = str(values_raw[expected_column_labels[column_id]])
            ammount = Decimal(ammount.strip())
            assert isinstance(ammount, Decimal)

            # requisition number --> IntegerField
            column_id = 4
            reqnum = int(values_raw[expected_column_labels[column_id]])
            try:
                order = Order.objects.get(order_reqnum=reqnum)
            except Order.DoesNotExist:
                # requisition number is valid, proceed
                pass
            else:
                if order.responsible.username in ["teresa.dias", "simone.zacarias"]:
                    msg = " ".join(
                        [
                            self.style.WARNING(f"Row {row+1:3d}:"),
                            f"Ignoring duplicated Order",
                        ]
                    )
                    self.stdout.write(msg)
                else:
                    msg = " ".join(
                        [
                            self.style.ERROR(f"Row {row+1:3d}:"),
                            f"Requisiton number {reqnum} already exists",
                        ]
                    )
                    self.stdout.write(msg)
                continue
            assert isinstance(reqnum, int)

            # requisition date --> DateField
            column_id = 5
            req_date = values_raw[expected_column_labels[column_id]]  # as float
            req_date = xlrd.xldate_as_tuple(req_date, workbook.datemode)
            req_date = datetime.datetime(*req_date).date()
            assert isinstance(req_date, datetime.date)

            # purchase order number --> CharField
            column_id = 6
            ponum = values_raw[expected_column_labels[column_id]]
            if ponum != "":
                try:
                    OrderExpenseCode.objects.get(purchase_order=ponum)
                except OrderExpenseCode.DoesNotExist:
                    assert ponum.startswith("EC")
                    if "," in ponum or ";" in ponum:
                        msg = " ".join(
                            [
                                self.style.WARNING(f"Row {row+1:3d}:"),
                                f"Using only the first Purchase Order number '{ponum}'",
                            ]
                        )
                        self.stdout.write(msg)
                        ponum = ponum.split(",")[0].split(";")[0]
                except OrderExpenseCode.MultipleObjectsReturned:
                    msg = " ".join(
                        [
                            self.style.WARNING(f"Row {row+1:3d}:"),
                            f"Multiple duplicated Purchase Order '{ponum}'",
                        ]
                    )
                    self.stdout.write(msg)
                else:
                    msg = " ".join(
                        [
                            self.style.WARNING(f"Row {row+1:3d}:"),
                            f"Duplicated Purchase Order '{ponum}'",
                        ]
                    )
                    self.stdout.write(msg)
            assert isinstance(ponum, str)

            # expense code --> ForeignKey to OrderExpenseCode
            column_id = 7
            expense_code_type = values_raw[expected_column_labels[column_id]]
            expense_code_type = expense_code_type.replace("&", "and")  # special case
            assert isinstance(expense_code_type, str)

            # cost center --> ForeignKey to FinanceCostCenter
            column_id = 8
            cost_center_code = int(values_raw[expected_column_labels[column_id]])
            cost_center_code = f"{cost_center_code:07d}"
            assert isinstance(cost_center_code, str)

            # finance project --> ForeignKey to FinanceProject
            column_id = 9
            finance_project_code = int(values_raw[expected_column_labels[column_id]])
            finance_project_code = f"{finance_project_code:03d}"
            assert isinstance(finance_project_code, str)

            # group --> ForeignKey to auth.Group
            column_id = 10
            auth_group_name = values_raw[expected_column_labels[column_id]]
            assert "\n" not in auth_group_name, "Invalid group name"
            assert isinstance(auth_group_name, str)
            if auth_group_name == "INDP":
                auth_group_name += " Lab"
            auth_group_name = f"GROUP: {auth_group_name}"
            try:
                group = AuthGroup.objects.get(name=auth_group_name.strip())
            except AuthGroup.DoesNotExist:
                raise CommandError("Invalid group name %s" % auth_group_name)
            assert isinstance(group, AuthGroup)

            # payment method --> CharField
            column_id = 11
            payment_method = values_raw[expected_column_labels[column_id]]
            for (
                payment_method_code,
                payment_method_name,
            ) in Order.PAYMENT_METHOD.items():
                if payment_method_name == payment_method.strip():
                    break
            else:
                raise CommandError("Payment Method mismatch: %s" % payment_method)
            payment_method = payment_method_code

            # notes --> TextField
            column_id = 12
            notes = values_raw[expected_column_labels[column_id]]
            assert isinstance(notes, str)

            # responsible --> (will be Simone by default)
            responsible = User.objects.get(username="simone.zacarias")

            # INTERLUDE: fetch the related tables required and build OrderExpenseCode
            try:
                cost_center = FinanceCostCenter.objects.get(
                    costcenter_code=cost_center_code
                )
            except FinanceCostCenter.DoesNotExist:
                raise CommandError("Invalid cost center code: %s" % cost_center_code)

            try:
                finance_project = FinanceProject.objects.get(
                    financeproject_code=finance_project_code, costcenter=cost_center
                )
            except FinanceProject.DoesNotExist:
                raise CommandError("Invalid project code: %s" % finance_project_code)

            try:
                expense_code = ExpenseCode.objects.get(
                    expensecode_type=expense_code_type, financeproject=finance_project
                )
            except ExpenseCode.DoesNotExist:
                raise CommandError("Invalid expense code: %s" % expense_code_type)

            # create Order
            order = Order(
                order_reqnum=reqnum,
                order_reqdate=req_date,
                order_desc=description,
                order_amount=ammount,
                order_req=requester_name,
                order_paymethod=payment_method,
                order_notes=notes,
                supplier=supplier,
                responsible=responsible,
                currency=Currency.objects.get(pk=1),  # EUR
                group=group,
            )
            order.save(
                expensecode_kwargs={
                    "expensecode": expense_code,
                    "purchase_order": ponum,
                }
            )

            imported_orders.append(order)

        self.stdout.write(
            self.style.SUCCESS("Imported %d orders" % len(imported_orders))
        )



from django.db import models
from .costcenter_queryset import CostCenterQuerySet


class CostCenter(models.Model):
    """
    Represents a Finance cost center in the system
    Example: CCU
    """
    name       = models.CharField('Name', max_length=200)  #: Name
    code       = models.CharField('Code', max_length=200)  #: Code
    start_date = models.DateField(null=True, blank=True)
    end_date   = models.DateField(null=True, blank=True)

    group      = models.ForeignKey('people.Group', verbose_name='Group', blank=False, null=False, on_delete=models.CASCADE)

    objects = CostCenterQuerySet.as_manager()

    class Meta:
        ordering            = ['code',]
        verbose_name        = "Finance Cost Center"
        verbose_name_plural = "Finance Cost Centers"

    def __str__(self):
        return "{0} - {1}".format(self.code, self.name)



import urllib

import xlrd
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.safestring import mark_safe




def build_missing_related_object_message(model, obj_repr, msg='', *args, **kwargs):
    """Returns a safe HTML message indicating that the parsed `obj_repr`
    does not exist in the database and provides a link to its add form.

    Extra `args` are used to format a custom message. Passing `kwargs`
    will populate the add form with initial data.

    Example::

        build_missing_related_object_message(
            FinanceCostCenter,
            '123456 - Super Cost Center',
            costcenter_code=cost_center_code,
            costcenter_name=cost_center_name,
        )

    Example with a custom message (notice that `model` and `obj_repr`
    are available for formatting)::

        build_missing_related_object_message(
            Budget,
            None,
            '{model._meta.verbose_name} for year {} with Expense Code {} does not exist',
            self.year,
            expense_code.abbrv,
            budget_year=self.year,
            budget_amount=amount,
            expensecode=expense_code.pk,
        )
    """

    if msg:
        msg = msg.format(*args, model=model, obj_repr=obj_repr)
    else:
        msg = f'{model._meta.verbose_name} "{obj_repr}" does not exist.'
    # TODO build url for pyforms app
    url = reverse(f'admin:{model._meta.app_label}_{model._meta.model_name}_add')
    url = url + '?' + urllib.parse.urlencode(kwargs)
    link = f'<a href="{url}" target="_blank">Click to add</a>'
    return mark_safe(msg + ' ' + link)





class BudgetDoc(models.Model):

    document = models.FileField(upload_to='documents/%Y/%m/%d')
    year     = models.IntegerField('Year')

    class Meta:
        verbose_name = 'Budget document'

    def __str__(self):
        return self.document.name

    def save(self):
        super().save()

        title = f'The following budgets for the year {self.year} were updated:'
        entry_template = '<li>{}: new amount is {:.2f} \u20AC</li>'

        update_report = title

        for budget, amount in self.budgets_to_update:
            budget.amount = amount
            budget.save()

            update_report += entry_template.format(
                budget.expensecode.abbrv, amount)

        self.update_report = mark_safe(update_report)



    def clean(self):
        """
        Custom clean method to verify the uploaded file is valid and that
        the objects required are already present in the database.
        """

        if not self.document:
            # skip file content validation if no file was selected
            return

        from finance.models import Budget
        from finance.models import CostCenter
        from finance.models import Project
        from finance.models import ExpenseCode

        budgets_to_update = []
        errors = []

        workbook = xlrd.open_workbook(
            file_contents=self.document.read(),
            encoding_override='utf-8',
            on_demand=True,
        )
        worksheet = workbook.sheet_by_index(0)

        # start by validating the file structure
        #   - number, name and order of columns
        #   - empty cells

        expected_column_labels = [
            'Projecto',
            'Centro Responsabilidade',
            'Grupo Despesa',
            'Data',
            'Valor',
        ]
        expected_column_number = len(expected_column_labels)

        if worksheet.ncols != expected_column_number:
            raise ValidationError(
                'Found %(found)s columns, expected exactly %(expected)s',
                params={
                    'found': worksheet.ncols,
                    'expected': expected_column_number,
                }
            )

        for col, expected in zip(range(worksheet.ncols), expected_column_labels):
            if worksheet.cell_value(0, col) != expected:
                errors.append('Column "%s" not found' % expected)

        if errors:
            raise ValidationError(errors)

        for col in range(worksheet.ncols):
            col_values = worksheet.col_values(col, start_rowx=2)
            empty_cells_row_indices = [
                i for i, value in enumerate(col_values) if value == ''
            ]
            for row in empty_cells_row_indices:
                cell = xlrd.formula.cellname(row, col)
                errors.append('Missing value in cell %s' % cell)

        if errors:
            raise ValidationError(errors)

        # validate columns values
        # due to model relationships the validation order must be:
        #   1. cost center
        #   2. finance project (requires cost center)
        #   3. expense code (required project)
        #   4. budget (requires expense code)


        for row in range(2, worksheet.nrows):
            values = worksheet.row_values(row)

            # replace en-dash for single dash across all str values extracted
            values = [
                value.replace(u"\u2013", '-')
                if isinstance(value, str) else value
                for value in values
            ]

            project_raw, cost_center_raw, expense_code_raw, _, amount = values

            expense_code = None

            # 1. cost center

            cost_center_code, cost_center_name = cost_center_raw.split(' - ', 1)
            cost_center_code = cost_center_code.zfill(7)
            try:
                cost_center = CostCenter.objects.get(
                    costcenter_code=cost_center_code,
                )
            except CostCenter.DoesNotExist:
                errors.append(
                    build_missing_related_object_message(
                        CostCenter,
                        cost_center_raw,
                        costcenter_code=cost_center_code,
                        costcenter_name=cost_center_name,
                    )
                )
                continue

            # 2. finance project

            project_code, project_name = project_raw.split(' - ', 1)
            project_code = project_code.zfill(3)
            try:
                project = Project.objects.get(
                    financeproject_code=project_code,
                    costcenter=cost_center,
                )
            except Project.DoesNotExist:
                errors.append(
                    build_missing_related_object_message(
                        Project,
                        project_raw,
                        financeproject_code=project_code,
                        financeproject_name=project_name,
                        costcenter=cost_center.pk,
                    )
                )
                continue

            # 3. expense code

            expense_code_number, expense_code_name = expense_code_raw.split(' - ', 1)
            expense_code_number = expense_code_number.zfill(2)
            try:
                expense_code = ExpenseCode.objects.get(
                    expensecode_number=expense_code_number,
                    expensecode_type=expense_code_name,
                    financeproject=project,
                )
            except ExpenseCode.DoesNotExist:
                errors.append(
                    build_missing_related_object_message(
                        ExpenseCode,
                        expense_code_raw,
                        expensecode_number=expense_code_number,
                        expensecode_type=expense_code_name,
                        financeproject=project.pk,
                    )
                )
                continue

            # 4. budget

            try:
                budget = Budget.objects.get(
                    expensecode=expense_code,
                    year=self.year,
                )
            except Budget.DoesNotExist:
                errors.append(
                    build_missing_related_object_message(
                        Budget,
                        None,
                        '{model._meta.verbose_name} for year {} with Expense Code {} does not exist',
                        self.year,
                        expense_code.abbrv,
                        year=self.year,
                        amount=amount,
                        expensecode=expense_code.pk,
                    )
                )
                continue

            budgets_to_update.append(
                (budget, amount)
            )

        if errors:
            raise ValidationError(list(set(errors)))

        self.budgets_to_update = budgets_to_update




from django.db import models

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


class Budget(models.Model):
    """
    Represents a Budgets in the system
    """

    amount = models.DecimalField('Amount (NET)', max_digits=11, decimal_places=2)
    year   = models.IntegerField('Year', blank=True, null=True)

    expensecode = models.ForeignKey('ExpenseCode', verbose_name='Expense Code', on_delete=models.CASCADE)

    class Meta:
        verbose_name = "Budget"
        verbose_name_plural = "Budgets"
        unique_together = ( ('year','expensecode'),)

    def __str__(self):
        return self.amount



from .budget.budget import Budget
from .budget.budget_doc import BudgetDoc

from .grant.grant import Grant
from .grant.financing_agency import FinancingAgency
from .grant.grant_domain import Grantdomain

from .costcenter.costcenter import CostCenter
from .project.project import Project
from .project.expense_code import ExpenseCode
